#include "cdstd.h"
#include <UnitTest++.h>
#include "cddefines.h"
#include "thirdparty.h"
#include "vectorize.h"

namespace {

	TEST(TestAvxPool)
	{
		// there is not a lot we can check here since the interesting bits are hidden
		// behind the interface, so the real testing will be done with valgrind-test
		t_avx_pool pool;
		init_genrand( 376356017L );
		for( int i=0; i < 2048; ++i )
		{
			size_t sz = genrand_int32()%200000;
			char *p = static_cast<char*>(pool.avx_alloc(sz));
			size_t ip = reinterpret_cast<size_t>(p);
			// test for correct alignment
			CHECK( p != NULL && (ip/CD_ALIGN)*CD_ALIGN == ip );
			for( size_t j=0; j < sz; ++j )
				p[j] = 'c';
			pool.avx_free(p);
		}
		// check that many simultaneous requests work correctly
		char* pp[50];
		for( int i=0; i < 50; ++i )
		{
			pp[i] = static_cast<char*>(pool.avx_alloc(4000));
			size_t ip = reinterpret_cast<size_t>(pp[i]);
			CHECK( pp[i] != NULL && (ip/CD_ALIGN)*CD_ALIGN == ip );
			for( size_t j=0; j < 4000; ++j )
				pp[i][j] = 'c';
		}
		for( int i=0; i < 50; ++i )
			pool.avx_free(pp[i]);
	}

	TEST(TestAvxPtr)
	{
		avx_ptr<double,true> p1(100);
		CHECK( p1.data() != NULL && p1.data() == p1.ptr0() );
		for( int i=0; i < 100; ++i )
			p1[i] = 1.;
		CHECK_THROW(( p1[-1] = 1. ),out_of_range);
		CHECK_THROW(( p1[100] = 1. ),out_of_range);
		avx_ptr<double> p2(0);
		CHECK( p2.data() == NULL );
		avx_ptr<double> p3(-10);
		CHECK( p3.data() == NULL );
		avx_ptr<double,true> p4(-10,100);
		CHECK( p4.data() != NULL && p4.data() == p4.ptr0()-10 );
		for( int i=-10; i < 100; ++i )
			p4[i] = 1.;
		CHECK_THROW(( p4[-11] = 1. ),out_of_range);
		CHECK_THROW(( p1[100] = 1. ),out_of_range);
		avx_ptr<double> p5(10,10);
		CHECK( p5.data() == NULL );
		avx_ptr<double> p6(10,-10);
		CHECK( p6.data() == NULL );
	}

	TEST(TestAllocatorAvx)
	{
		// there is not a lot we can check here since the interesting bits are hidden
		// behind the interface, so the real testing will be done with valgrind-test
		init_genrand( 376833017L );
		for( int i=0; i < 2048; ++i )
		{
			size_t sz = genrand_int32()%200000;
			vector< int, allocator_avx<int> > a(sz);
			size_t ip = reinterpret_cast<size_t>(get_ptr(a));
			CHECK( ip != 0 && (ip/CD_ALIGN)*CD_ALIGN == ip );
			for( size_t j=0; j < sz; ++j )
				a[j] = 3;
		}
	}

	TEST(TestReduceAd)
	{
		double a[32];
		for( int i=0; i < 32; ++i )
			a[i] = double(i);
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_a( a, ilo, ihi );
				double expect = double(ihi*(ihi-1)/2 - ilo*(ilo-1)/2);
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceAf)
	{
		sys_float a[32];
		for( int i=0; i < 32; ++i )
			a[i] = sys_float(i);
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				sys_float val = reduce_a( a, ilo, ihi );
				sys_float expect = sys_float(ihi*(ihi-1)/2 - ilo*(ilo-1)/2);
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABd)
	{
		double a[32], b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_ab( a, b, ilo, ihi );
				double expect = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABdf)
	{
		double a[32];
		sys_float b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_ab( a, b, ilo, ihi );
				double expect = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABf)
	{
		sys_float a[32], b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = sys_float(i);
			b[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				sys_float val = reduce_ab( a, b, ilo, ihi );
				sys_float expect = sys_float(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABCd)
	{
		double a[32], b[32], c[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
			c[i] = double(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_abc( a, b, c, ilo, ihi );
				double expect = double(pow2(ihi*(ihi-1)/2) - pow2(ilo*(ilo-1)/2));
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABCddf)
	{
		double a[32], b[32];
		sys_float c[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
			c[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_abc( a, b, c, ilo, ihi );
				double expect = double(pow2(ihi*(ihi-1)/2) - pow2(ilo*(ilo-1)/2));
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABCdff)
	{
		double a[32];
		sys_float b[32], c[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = sys_float(i);
			c[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val = reduce_abc( a, b, c, ilo, ihi );
				double expect = double(pow2(ihi*(ihi-1)/2) - pow2(ilo*(ilo-1)/2));
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceABCf)
	{
		sys_float a[32], b[32], c[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = sys_float(i);
			b[i] = sys_float(i);
			c[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				sys_float val = reduce_abc( a, b, c, ilo, ihi );
				sys_float expect = sys_float(pow2(ihi*(ihi-1)/2) - pow2(ilo*(ilo-1)/2));
				CHECK( fp_equal( val, expect ) );
			}
		}
	}

	TEST(TestReduceAB_Ad)
	{
		double a[32], b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val2, val = reduce_ab_a( a, b, ilo, ihi, &val2 );
				double expect = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				double expect2 = double(ihi*(ihi-1)/2 - ilo*(ilo-1)/2);
				CHECK( fp_equal( val, expect ) );
				CHECK( fp_equal( val2, expect2 ) );
			}
		}
	}

	TEST(TestReduceAB_Adf)
	{
		double a[32];
		sys_float b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = sys_float(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val2, val = reduce_ab_a( a, b, ilo, ihi, &val2 );
				double expect = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				double expect2 = double(ihi*(ihi-1)/2 - ilo*(ilo-1)/2);
				CHECK( fp_equal( val, expect ) );
				CHECK( fp_equal( val2, expect2 ) );
				val = reduce_ab_a( b, a, ilo, ihi, &val2 );
				CHECK( fp_equal( val, expect ) );
				CHECK( fp_equal( val2, expect2 ) );
			}
		}
	}

	TEST(TestReduceAB_Af)
	{
		double a[32], b[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val2, val = reduce_ab_a( a, b, ilo, ihi, &val2 );
				double expect = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				double expect2 = double(ihi*(ihi-1)/2 - ilo*(ilo-1)/2);
				CHECK( fp_equal( val, expect ) );
				CHECK( fp_equal( val2, expect2 ) );
			}
		}
	}

	TEST(TestReduceABC_ABd)
	{
		double a[32], b[32], c[32];
		for( int i=0; i < 32; ++i )
		{
			a[i] = double(i);
			b[i] = double(i);
			c[i] = double(i);
		}
		for( int ilo=0; ilo < 32; ilo++ )
		{
			for( int ihi=ilo+1; ihi <= 32; ihi++ )
			{
				double val2, val = reduce_abc_ab( a, b, c, ilo, ihi, &val2 );
				double expect = double(pow2(ihi*(ihi-1)/2) - pow2(ilo*(ilo-1)/2));
				double expect2 = double(ihi*(ihi-1)*(2*ihi-1)/6 - ilo*(ilo-1)*(2*ilo-1)/6);
				CHECK( fp_equal( val, expect ) );
				CHECK( fp_equal( val2, expect2 ) );
			}
		}
	}

	TEST(TestExp10d)
	{
		// exp10(double) is defined in cddefines.h
		double ln10 = log(10.);
		double maxarg = log10(DBL_MAX);
		init_genrand( 392916017L );
		for( int i=0; i < 2048; ++i )
		{
			double arg = (genrand_real1()-0.5)*617.;
			if( arg < maxarg )
			{
				double val1 = exp10(arg);
				double val2 = pow(10.,arg);
				CHECK( fp_equal( val1, val2, max((int)abs(arg*ln10),3) ) );
			}
		}
	}

	TEST(TestExp10f)
	{
		// exp10(sys_float) is defined in cddefines.h
		sys_float ln10 = logf(10.f);
		sys_float maxarg = log10f(FLT_MAX);
		init_genrand( 392977017L );
		for( int i=0; i < 2048; ++i )
		{
			sys_float arg = (genrand_real1()-0.5)*78.f;
			if( arg < maxarg )
			{
				sys_float val1 = exp10(arg);
				sys_float val2 = powf(10.f,arg);
				CHECK( fp_equal( val1, val2, max((int)abs(arg*ln10),3) ) );
			}
		}
	}

	TEST(TestVexpd)
	{
		const int sz = 2048;
		ALIGNED(CD_ALIGN) double arg[sz];
		ALIGNED(CD_ALIGN) double val[sz];
#ifdef __AVX__
		double minarg = log(DBL_MIN);
		// check correct behavior near the underflow margin
		// our implementation doesn't support gradual underflow
		arg[4] = minarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafter(arg[i+1],-DBL_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafter(arg[i-1],DBL_MAX);
		// the alignment above guarantees that the vectorized routine is used for all arguments
		vexp( arg, val, 0, 8 );
		for( int i=0; i < 4; ++i )
			CHECK( val[i] == 0. );
		// don't test val[4]: rounding may have gone either way
		// so we don't know what to expect...
		for( int i=5; i < 8; ++i )
			CHECK( fp_equal( val[i], exp(arg[i]) ) );
		// now do the same thing near the overflow margin
		double maxarg = log(DBL_MAX);
		arg[4] = maxarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafter(arg[i+1],-DBL_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafter(arg[i-1],DBL_MAX);
		vexp( arg, val, 0, 8 );
		for( int i=0; i < 4; ++i )
			CHECK( fp_equal( val[i], exp(arg[i]) ) );
		for( int i=5; i < 8; ++i )
			CHECK( isinf(val[i]) );
		// now check results over the entire range
		init_genrand( 395256017L );
		for( int i=0; i < sz; ++i )
			arg[i] = (genrand_real1()-0.5)*1420.;
		vexp( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
		{
			if( arg[i] < minarg )
				CHECK( val[i] == 0. );
			else if( arg[i] > maxarg )
				CHECK( isinf(val[i]) );
			else
				CHECK( fp_equal( val[i], exp(arg[i]) ) );
		}
#endif
		for( int i=0; i < 32; ++i )
			arg[i] = double(i);
		// finally check that non-aligned arrays are treated correctly
		for( int i=0; i < 32; ++i )
		{
			for( int j=i+1; j <= 32; ++j )
			{
				invalidate_array( val, 32*sizeof(double) );
				vexp( arg, val, i, j );
				for( int k=0; k < 32; ++k )
				{
					if( k < i || k >= j )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], exp(arg[k]) ) );
				}
				invalidate_array( val, 32*sizeof(double) );
				vexp( &arg[i], val, 0, j-i );
				for( int k=0; k < 32; ++k )
				{
					if( k >= j-i )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], exp(arg[k+i]) ) );
				}
			}
		}
	}

	TEST(TestVexp10d)
	{
		const int sz = 2048;
		ALIGNED(CD_ALIGN) double arg[sz];
		ALIGNED(CD_ALIGN) double val[sz];
		double ln10 = log(10.);
#ifdef __AVX__
		double minarg = log10(DBL_MIN);
		arg[4] = minarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafter(arg[i+1],-DBL_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafter(arg[i-1],DBL_MAX);
		vexp10( arg, val, 0, 8 );
		for( int i=0; i < 3; ++i )
			CHECK( val[i] == 0. );
		// for vexp10 we don't test val[3] through val[5] since multiplication
		// with log(10.) increases round-off error in the argument of vexp10
		// for the same reason we also need an extended tolerance in fp_equal()
		for( int i=6; i < 8; ++i )
			CHECK( fp_equal( val[i], pow(10.,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		double maxarg = log10(DBL_MAX);
		arg[4] = maxarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafter(arg[i+1],-DBL_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafter(arg[i-1],DBL_MAX);
		vexp10( arg, val, 0, 8 );
		for( int i=0; i < 3; ++i )
			CHECK( fp_equal( val[i], pow(10.,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		for( int i=6; i < 8; ++i )
			CHECK( isinf(val[i]) );
		init_genrand( 395316017L );
		for( int i=0; i < sz; ++i )
			arg[i] = (genrand_real1()-0.5)*617.;
		vexp10( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
		{
			if( arg[i] < minarg )
				CHECK( val[i] == 0. );
			else if( arg[i] > maxarg )
				CHECK( isinf(val[i]) );
			else
				CHECK( fp_equal( val[i], pow(10.,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		}
#endif
		for( int i=0; i < 32; ++i )
			arg[i] = double(i);
		for( int i=0; i < 32; ++i )
		{
			for( int j=i+1; j <= 32; ++j )
			{
				invalidate_array( val, 32*sizeof(double) );
				vexp10( arg, val, i, j );
				for( int k=0; k < 32; ++k )
				{
					if( k < i || k >= j )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], pow(10.,arg[k]),
								 max((int)abs(arg[k]*ln10),3) ) );
				}
				invalidate_array( val, 32*sizeof(double) );
				vexp10( &arg[i], val, 0, j-i );
				for( int k=0; k < 32; ++k )
				{
					if( k >= j-i )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], pow(10.,arg[k+i]),
								 max((int)abs(arg[k+i]*ln10),3) ) );
				}
			}
		}
	}

	TEST(TestVexpf)
	{
		const int sz = 2048;
		ALIGNED(CD_ALIGN) sys_float arg[sz];
		ALIGNED(CD_ALIGN) sys_float val[sz];
#ifdef __AVX__
		sys_float minarg = logf(FLT_MIN);
		arg[4] = minarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafterf(arg[i+1],-FLT_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafterf(arg[i-1],FLT_MAX);
		vexp( arg, val, 0, 8 );
		for( int i=0; i < 4; ++i )
			CHECK( val[i] == 0.f );
		for( int i=5; i < 8; ++i )
			CHECK( fp_equal( val[i], expf(arg[i]) ) );
		sys_float maxarg = logf(FLT_MAX);
		arg[4] = maxarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafterf(arg[i+1],-FLT_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafterf(arg[i-1],FLT_MAX);
		vexp( arg, val, 0, 8 );
		for( int i=0; i < 4; ++i )
			CHECK( fp_equal( val[i], expf(arg[i]) ) );
		for( int i=5; i < 8; ++i )
			CHECK( isinff(val[i]) );
		init_genrand( 395288017L );
		for( int i=0; i < sz; ++i )
			arg[i] = sys_float(genrand_real1()-0.5)*178.f;
		vexp( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
		{
			if( arg[i] < minarg )
				CHECK( val[i] == 0.f );
			else if( arg[i] > maxarg )
				CHECK( isinff(val[i]) );
			else
				CHECK( fp_equal( val[i], expf(arg[i]) ) );
		}
#endif
		for( int i=0; i < 32; ++i )
			arg[i] = sys_float(i);
		for( int i=0; i < 32; ++i )
		{
			for( int j=i+1; j <= 32; ++j )
			{
				invalidate_array( val, 32*sizeof(sys_float) );
				vexp( arg, val, i, j );
				for( int k=0; k < 32; ++k )
				{
					if( k < i || k >= j )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], expf(arg[k]) ) );
				}
				invalidate_array( val, 32*sizeof(sys_float) );
				vexp( &arg[i], val, 0, j-i );
				for( int k=0; k < 32; ++k )
				{
					if( k >= j-i )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], expf(arg[k+i]) ) );
				}
			}
		}
	}

	TEST(TestVexp10f)
	{
		const int sz = 2048;
		ALIGNED(CD_ALIGN) sys_float arg[sz];
		ALIGNED(CD_ALIGN) sys_float val[sz];
		sys_float ln10 = logf(10.f);
#ifdef __AVX__
		sys_float minarg = log10f(FLT_MIN);
		arg[4] = minarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafterf(arg[i+1],-FLT_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafterf(arg[i-1],FLT_MAX);
		vexp10( arg, val, 0, 8 );
		for( int i=0; i < 3; ++i )
			CHECK( val[i] == 0.f );
		for( int i=6; i < 8; ++i )
			CHECK( fp_equal( val[i], powf(10.f,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		sys_float maxarg = log10f(FLT_MAX);
		arg[4] = maxarg;
		for( int i=3; i >= 0; --i )
			arg[i] = nextafterf(arg[i+1],-FLT_MAX);
		for( int i=5; i < 8; ++i )
			arg[i] = nextafterf(arg[i-1],FLT_MAX);
		vexp10( arg, val, 0, 8 );
		for( int i=0; i < 3; ++i )
			CHECK( fp_equal( val[i], powf(10.f,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		for( int i=6; i < 8; ++i )
			CHECK( isinff(val[i]) );
		init_genrand( 395313517L );
		for( int i=0; i < sz; ++i )
			arg[i] = sys_float(genrand_real1()-0.5)*78.f;
		vexp10( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
		{
			if( arg[i] < minarg )
				CHECK( val[i] == 0.f );
			else if( arg[i] > maxarg )
				CHECK( isinff(val[i]) );
			else
				CHECK( fp_equal( val[i], powf(10.f,arg[i]), max((int)abs(arg[i]*ln10),3) ) );
		}
#endif
		for( int i=0; i < 32; ++i )
			arg[i] = sys_float(i);
		for( int i=0; i < 32; ++i )
		{
			for( int j=i+1; j <= 32; ++j )
			{
				invalidate_array( val, 32*sizeof(sys_float) );
				vexp10( arg, val, i, j );
				for( int k=0; k < 32; ++k )
				{
					if( k < i || k >= j )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], powf(10.f,arg[k]),
								 max((int)abs(arg[k]*ln10),3) ) );
				}
				invalidate_array( val, 32*sizeof(sys_float) );
				vexp10( &arg[i], val, 0, j-i );
				for( int k=0; k < 32; ++k )
				{
					if( k >= j-i )
						CHECK( isnan(val[k]) );
					else
						CHECK( fp_equal( val[k], powf(10.f,arg[k+i]),
								 max((int)abs(arg[k+i]*ln10),3) ) );
				}
			}
		}
	}

	TEST(TestVexpdx4)
	{
		double y[4];
		vexp(y,1.1,1.2,1.3,1.4);
		CHECK( fp_equal( y[0], exp(1.1) ) );
		CHECK( fp_equal( y[1], exp(1.2) ) );
		CHECK( fp_equal( y[2], exp(1.3) ) );
		CHECK( fp_equal( y[3], exp(1.4) ) );
	}

	TEST(TestVexp10dx4)
	{
		double y[4];
		vexp10(y,1.1,1.2,1.3,1.4);
		CHECK( fp_equal( y[0], exp10(1.1) ) );
		CHECK( fp_equal( y[1], exp10(1.2) ) );
		CHECK( fp_equal( y[2], exp10(1.3) ) );
		CHECK( fp_equal( y[3], exp10(1.4) ) );
	}

	TEST(TestVexpdx8)
	{
		double y[8];
		vexp(y,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8);
		CHECK( fp_equal( y[0], exp(1.1) ) );
		CHECK( fp_equal( y[1], exp(1.2) ) );
		CHECK( fp_equal( y[2], exp(1.3) ) );
		CHECK( fp_equal( y[3], exp(1.4) ) );
		CHECK( fp_equal( y[4], exp(1.5) ) );
		CHECK( fp_equal( y[5], exp(1.6) ) );
		CHECK( fp_equal( y[6], exp(1.7) ) );
		CHECK( fp_equal( y[7], exp(1.8) ) );
	}

	TEST(TestVexp10dx8)
	{
		double y[8];
		vexp10(y,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8);
		CHECK( fp_equal( y[0], exp10(1.1) ) );
		CHECK( fp_equal( y[1], exp10(1.2) ) );
		CHECK( fp_equal( y[2], exp10(1.3) ) );
		CHECK( fp_equal( y[3], exp10(1.4) ) );
		CHECK( fp_equal( y[4], exp10(1.5) ) );
		CHECK( fp_equal( y[5], exp10(1.6) ) );
		CHECK( fp_equal( y[6], exp10(1.7) ) );
		CHECK( fp_equal( y[7], exp10(1.8) ) );
	}

	TEST(TestVexpfx4)
	{
		sys_float y[4];
		vexp(y,1.1f,1.2f,1.3f,1.4f);
		CHECK( fp_equal( y[0], expf(1.1f) ) );
		CHECK( fp_equal( y[1], expf(1.2f) ) );
		CHECK( fp_equal( y[2], expf(1.3f) ) );
		CHECK( fp_equal( y[3], expf(1.4f) ) );
	}

	TEST(TestVexp10fx4)
	{
		sys_float y[4];
		vexp10(y,1.1f,1.2f,1.3f,1.4f);
		CHECK( fp_equal( y[0], exp10f(1.1f) ) );
		CHECK( fp_equal( y[1], exp10f(1.2f) ) );
		CHECK( fp_equal( y[2], exp10f(1.3f) ) );
		CHECK( fp_equal( y[3], exp10f(1.4f) ) );
	}

	TEST(TestVexpfx8)
	{
		sys_float y[8];
		vexp(y,1.1f,1.2f,1.3f,1.4f,1.5f,1.6f,1.7f,1.8f);
		CHECK( fp_equal( y[0], expf(1.1f) ) );
		CHECK( fp_equal( y[1], expf(1.2f) ) );
		CHECK( fp_equal( y[2], expf(1.3f) ) );
		CHECK( fp_equal( y[3], expf(1.4f) ) );
		CHECK( fp_equal( y[4], expf(1.5f) ) );
		CHECK( fp_equal( y[5], expf(1.6f) ) );
		CHECK( fp_equal( y[6], expf(1.7f) ) );
		CHECK( fp_equal( y[7], expf(1.8f) ) );
	}

	TEST(TestVexp10fx8)
	{
		sys_float y[8];
		vexp10(y,1.1f,1.2f,1.3f,1.4f,1.5f,1.6f,1.7f,1.8f);
		CHECK( fp_equal( y[0], exp10f(1.1f) ) );
		CHECK( fp_equal( y[1], exp10f(1.2f) ) );
		CHECK( fp_equal( y[2], exp10f(1.3f) ) );
		CHECK( fp_equal( y[3], exp10f(1.4f) ) );
		CHECK( fp_equal( y[4], exp10f(1.5f) ) );
		CHECK( fp_equal( y[5], exp10f(1.6f) ) );
		CHECK( fp_equal( y[6], exp10f(1.7f) ) );
		CHECK( fp_equal( y[7], exp10f(1.8f) ) );
	}

	TEST(TestVexpfx16)
	{
		sys_float y[16];
		vexp(y,1.1f,1.2f,1.3f,1.4f,1.5f,1.6f,1.7f,1.8f,2.11f,2.22f,2.33f,2.44f,2.55f,2.66f,2.77f,2.88f);
		CHECK( fp_equal( y[0], expf(1.1f) ) );
		CHECK( fp_equal( y[1], expf(1.2f) ) );
		CHECK( fp_equal( y[2], expf(1.3f) ) );
		CHECK( fp_equal( y[3], expf(1.4f) ) );
		CHECK( fp_equal( y[4], expf(1.5f) ) );
		CHECK( fp_equal( y[5], expf(1.6f) ) );
		CHECK( fp_equal( y[6], expf(1.7f) ) );
		CHECK( fp_equal( y[7], expf(1.8f) ) );
		CHECK( fp_equal( y[8], expf(2.11f) ) );
		CHECK( fp_equal( y[9], expf(2.22f) ) );
		CHECK( fp_equal( y[10], expf(2.33f) ) );
		CHECK( fp_equal( y[11], expf(2.44f) ) );
		CHECK( fp_equal( y[12], expf(2.55f) ) );
		CHECK( fp_equal( y[13], expf(2.66f) ) );
		CHECK( fp_equal( y[14], expf(2.77f) ) );
		CHECK( fp_equal( y[15], expf(2.88f) ) );
	}

	TEST(TestVexp10fx16)
	{
		sys_float y[16];
		vexp10(y,1.1f,1.2f,1.3f,1.4f,1.5f,1.6f,1.7f,1.8f,2.11f,2.22f,2.33f,2.44f,2.55f,2.66f,2.77f,2.88f);
		CHECK( fp_equal( y[0], exp10f(1.1f) ) );
		CHECK( fp_equal( y[1], exp10f(1.2f) ) );
		CHECK( fp_equal( y[2], exp10f(1.3f) ) );
		CHECK( fp_equal( y[3], exp10f(1.4f) ) );
		CHECK( fp_equal( y[4], exp10f(1.5f) ) );
		CHECK( fp_equal( y[5], exp10f(1.6f) ) );
		CHECK( fp_equal( y[6], exp10f(1.7f) ) );
		CHECK( fp_equal( y[7], exp10f(1.8f) ) );
		CHECK( fp_equal( y[8], exp10f(2.11f) ) );
		CHECK( fp_equal( y[9], exp10f(2.22f) ) );
		CHECK( fp_equal( y[10], exp10f(2.33f) ) );
		CHECK( fp_equal( y[11], exp10f(2.44f) ) );
		CHECK( fp_equal( y[12], exp10f(2.55f) ) );
		CHECK( fp_equal( y[13], exp10f(2.66f) ) );
		CHECK( fp_equal( y[14], exp10f(2.77f) ) );
		CHECK( fp_equal( y[15], exp10f(2.88f) ) );
	}

	TEST(TestVlogd)
	{
#ifdef __AVX__
		const int sz = 2048;
		ALIGNED(CD_ALIGN) double arg[sz];
		ALIGNED(CD_ALIGN) double val[sz];
		init_genrand( 395317717L );
		// this range includes subnormal numbers
		for( int i=0; i < sz; ++i )
			arg[i] = exp(genrand_real1()*1454.2-744.43);
		vlog( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
			CHECK( fp_equal( val[i], log(arg[i]) ) );
#else
		CHECK( fp_equal( 1., 1. ) );
#endif
	}

	TEST(TestVlog10d)
	{
#ifdef __AVX__
		const int sz = 2048;
		ALIGNED(CD_ALIGN) double arg[sz];
		ALIGNED(CD_ALIGN) double val[sz];
		init_genrand( 395288717L );
		// this range includes subnormal numbers
		for( int i=0; i < sz; ++i )
			arg[i] = exp10(genrand_real1()*631.5-323.27);
		vlog10( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
			CHECK( fp_equal( val[i], log10(arg[i]) ) );
#else
		CHECK( fp_equal( 1., 1. ) );
#endif
	}

	TEST(TestVlogf)
	{
#ifdef __AVX__
		const int sz = 2048;
		ALIGNED(CD_ALIGN) sys_float arg[sz];
		ALIGNED(CD_ALIGN) sys_float val[sz];
		init_genrand( 395955717L );
		// this range includes subnormal numbers
		for( int i=0; i < sz; ++i )
			arg[i] = exp(genrand_real1()*191.9-103.27);
		vlog( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
			CHECK( fp_equal( val[i], logf(arg[i]) ) );
#else
		CHECK( fp_equal( 1., 1. ) );
#endif
	}

	TEST(TestVlog10f)
	{
#ifdef __AVX__
		const int sz = 2048;
		ALIGNED(CD_ALIGN) sys_float arg[sz];
		ALIGNED(CD_ALIGN) sys_float val[sz];
		init_genrand( 395281237L );
		// this range includes subnormal numbers
		for( int i=0; i < sz; ++i )
			arg[i] = exp10(genrand_real1()*83.3-44.85);
		vlog10( arg, val, 0, sz );
		for( int i=0; i < sz; ++i )
			CHECK( fp_equal( val[i], log10f(arg[i]) ) );
#else
		CHECK( fp_equal( 1., 1. ) );
#endif
	}

	TEST(TestVlogdx4)
	{
		double y[4];
		vlog(y,11.,22.,33.,44.);
		CHECK( fp_equal( y[0], log(11.) ) );
		CHECK( fp_equal( y[1], log(22.) ) );
		CHECK( fp_equal( y[2], log(33.) ) );
		CHECK( fp_equal( y[3], log(44.) ) );
	}

	TEST(TestVlog10dx4)
	{
		double y[4];
		vlog10(y,11.,22.,33.,44.);
		CHECK( fp_equal( y[0], log10(11.) ) );
		CHECK( fp_equal( y[1], log10(22.) ) );
		CHECK( fp_equal( y[2], log10(33.) ) );
		CHECK( fp_equal( y[3], log10(44.) ) );
	}

	TEST(TestVlogdx8)
	{
		double y[8];
		vlog(y,11.,22.,33.,44.,55.,66.,77.,88.);
		CHECK( fp_equal( y[0], log(11.) ) );
		CHECK( fp_equal( y[1], log(22.) ) );
		CHECK( fp_equal( y[2], log(33.) ) );
		CHECK( fp_equal( y[3], log(44.) ) );
		CHECK( fp_equal( y[4], log(55.) ) );
		CHECK( fp_equal( y[5], log(66.) ) );
		CHECK( fp_equal( y[6], log(77.) ) );
		CHECK( fp_equal( y[7], log(88.) ) );
	}

	TEST(TestVlog10dx8)
	{
		double y[8];
		vlog10(y,11.,22.,33.,44.,55.,66.,77.,88.);
		CHECK( fp_equal( y[0], log10(11.) ) );
		CHECK( fp_equal( y[1], log10(22.) ) );
		CHECK( fp_equal( y[2], log10(33.) ) );
		CHECK( fp_equal( y[3], log10(44.) ) );
		CHECK( fp_equal( y[4], log10(55.) ) );
		CHECK( fp_equal( y[5], log10(66.) ) );
		CHECK( fp_equal( y[6], log10(77.) ) );
		CHECK( fp_equal( y[7], log10(88.) ) );
	}

	TEST(TestVlogfx4)
	{
		sys_float y[4];
		vlog(y,11.f,22.f,33.f,44.f);
		CHECK( fp_equal( y[0], logf(11.f) ) );
		CHECK( fp_equal( y[1], logf(22.f) ) );
		CHECK( fp_equal( y[2], logf(33.f) ) );
		CHECK( fp_equal( y[3], logf(44.f) ) );
	}

	TEST(TestVlog10fx4)
	{
		sys_float y[4];
		vlog10(y,11.f,22.f,33.f,44.f);
		CHECK( fp_equal( y[0], log10f(11.f) ) );
		CHECK( fp_equal( y[1], log10f(22.f) ) );
		CHECK( fp_equal( y[2], log10f(33.f) ) );
		CHECK( fp_equal( y[3], log10f(44.f) ) );
	}

	TEST(TestVlogfx8)
	{
		sys_float y[8];
		vlog(y,11.f,22.f,33.f,44.f,55.f,66.f,77.f,88.f);
		CHECK( fp_equal( y[0], logf(11.f) ) );
		CHECK( fp_equal( y[1], logf(22.f) ) );
		CHECK( fp_equal( y[2], logf(33.f) ) );
		CHECK( fp_equal( y[3], logf(44.f) ) );
		CHECK( fp_equal( y[4], logf(55.f) ) );
		CHECK( fp_equal( y[5], logf(66.f) ) );
		CHECK( fp_equal( y[6], logf(77.f) ) );
		CHECK( fp_equal( y[7], logf(88.f) ) );
	}

	TEST(TestVlog10fx8)
	{
		sys_float y[8];
		vlog10(y,11.f,22.f,33.f,44.f,55.f,66.f,77.f,88.f);
		CHECK( fp_equal( y[0], log10f(11.f) ) );
		CHECK( fp_equal( y[1], log10f(22.f) ) );
		CHECK( fp_equal( y[2], log10f(33.f) ) );
		CHECK( fp_equal( y[3], log10f(44.f) ) );
		CHECK( fp_equal( y[4], log10f(55.f) ) );
		CHECK( fp_equal( y[5], log10f(66.f) ) );
		CHECK( fp_equal( y[6], log10f(77.f) ) );
		CHECK( fp_equal( y[7], log10f(88.f) ) );
	}

	TEST(TestVlogfx16)
	{
		sys_float y[16];
		vlog(y,11.f,22.f,33.f,44.f,55.f,66.f,77.f,88.f,111.f,122.f,133.f,144.f,155.f,166.f,177.f,188.f);
		CHECK( fp_equal( y[0], logf(11.f) ) );
		CHECK( fp_equal( y[1], logf(22.f) ) );
		CHECK( fp_equal( y[2], logf(33.f) ) );
		CHECK( fp_equal( y[3], logf(44.f) ) );
		CHECK( fp_equal( y[4], logf(55.f) ) );
		CHECK( fp_equal( y[5], logf(66.f) ) );
		CHECK( fp_equal( y[6], logf(77.f) ) );
		CHECK( fp_equal( y[7], logf(88.f) ) );
		CHECK( fp_equal( y[8], logf(111.f) ) );
		CHECK( fp_equal( y[9], logf(122.f) ) );
		CHECK( fp_equal( y[10], logf(133.f) ) );
		CHECK( fp_equal( y[11], logf(144.f) ) );
		CHECK( fp_equal( y[12], logf(155.f) ) );
		CHECK( fp_equal( y[13], logf(166.f) ) );
		CHECK( fp_equal( y[14], logf(177.f) ) );
		CHECK( fp_equal( y[15], logf(188.f) ) );
	}

	TEST(TestVlog10fx16)
	{
		sys_float y[16];
		vlog10(y,11.f,22.f,33.f,44.f,55.f,66.f,77.f,88.f,111.f,122.f,133.f,144.f,155.f,166.f,177.f,188.f);
		CHECK( fp_equal( y[0], log10f(11.f) ) );
		CHECK( fp_equal( y[1], log10f(22.f) ) );
		CHECK( fp_equal( y[2], log10f(33.f) ) );
		CHECK( fp_equal( y[3], log10f(44.f) ) );
		CHECK( fp_equal( y[4], log10f(55.f) ) );
		CHECK( fp_equal( y[5], log10f(66.f) ) );
		CHECK( fp_equal( y[6], log10f(77.f) ) );
		CHECK( fp_equal( y[7], log10f(88.f) ) );
		CHECK( fp_equal( y[8], log10f(111.f) ) );
		CHECK( fp_equal( y[9], log10f(122.f) ) );
		CHECK( fp_equal( y[10], log10f(133.f) ) );
		CHECK( fp_equal( y[11], log10f(144.f) ) );
		CHECK( fp_equal( y[12], log10f(155.f) ) );
		CHECK( fp_equal( y[13], log10f(166.f) ) );
		CHECK( fp_equal( y[14], log10f(177.f) ) );
		CHECK( fp_equal( y[15], log10f(188.f) ) );
	}
}
