/* This file is part of Cloudy and is copyright (C)1978-2015 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
#include "cddefines.h"
#include "vectorize.h"
#include "vectorize_math.h"

//
// Written by Peter A.M. van Hoof, Royal Observatory of Belgium, Brussels
//
// this file contains vectorized versions of the single and double variants of the log()
// function. They are vectorized using AVX instructions, but also make use of AVX2, FMA,
// and AVX512 instructions when available. The basic algorithms for calculating the log()
// functions was slightly simplified from the openlibm library versions available at
// http://openlibm.org/ which is subject to the following copyright:
//
// ====================================================
// Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
//
// Developed at SunSoft, a Sun Microsystems, Inc. business.
// Permission to use, copy, modify, and distribute this
// software is freely granted, provided that this notice 
// is preserved.
// ====================================================
//
// This work was also inspired by similar routines written by Giovanni Garberoglio which
// are available at: http://software-lisc.fbk.eu/avx_mathfun/
//

#ifdef __AVX__

VECL_CONST(log_mask1,0x7fffffffffffffff);
#if defined(__AVX512F__) || defined(__AVX2__)
VECL_CONST(log_off,0x3ff);
VECL_CONST(log_mask2,0x000fffffffffffff);
VECL_CONST(log_sq2off,0x95f6400000000);
VECL_CONST(log_mask3,0x0010000000000000);
VECL_CONST(log_mask4,0x3ff0000000000000);
#else
VECI_CONST(log_off,0x3ff);
VECI_CONST(log_mask2,0x000fffff);
VECI_CONST(log_sq2off,0x95f64);
VECI_CONST(log_mask3,0x00100000);
VECI_CONST(log_mask4,0x3ff00000);
#endif

VECD_CONST(dbl_min,DBL_MIN);
VECD_CONST(two54,1.80143985094819840000e+16);
VECD_CONST(m54,-54.);
#ifdef __AVX512F__
VECD_CONST(mone,-1.);
#endif
VECD_CONST(zero,0.);
VECD_CONST(one,1.);
VECD_CONST(two,2.);

VECD_CONST(log_lg7,-1.479819860511658591e-01);
VECD_CONST(log_lg6,-1.531383769920937332e-01);
VECD_CONST(log_lg5,-1.818357216161805012e-01);
VECD_CONST(log_lg4,-2.222219843214978396e-01);
VECD_CONST(log_lg3,-2.857142874366239149e-01);
VECD_CONST(log_lg2,-3.999999999940941908e-01);
VECD_CONST(log_lg1,-6.666666666666735130e-01);

VECD_CONST(ln_two,log(2.));

VECD_CONST(ln_ten,log(10.));

#ifdef __AVX512F__

inline v8df v1logd(v8df x)
{
	v8df xarg = x;
	v8di ix = _mm512_castpd_si512(x);
	ix = _mm512_and_epi64(ix, log_mask1);
	x = _mm512_castsi512_pd(ix);
	__mmask8 snmask = _mm512_cmp_pd_mask(x, dbl_min, _CMP_LT_OS);
	x = _mm512_mask_mul_pd(x, snmask, x, two54);
	ix = _mm512_castpd_si512(x);
	v8di k = _mm512_srli_epi64(ix, 52);
	k = _mm512_sub_epi64(k, log_off);
	v8di iy = _mm512_and_epi64(ix, log_mask2);
	v8di iz = _mm512_add_epi64(iy, log_sq2off);
	iz = _mm512_and_epi64(iz, log_mask3);
	v8di ie = _mm512_xor_epi64(iz, log_mask4);
	iy = _mm512_or_epi64(iy, ie);
	iz = _mm512_srli_epi64(iz, 52);
	k = _mm512_add_epi64(k, iz);
	v8df f = _mm512_castsi512_pd(iy);
	f = _mm512_sub_pd(f, one);
	v8df s = _mm512_add_pd(f, two);
	s = _mm512_div_pd(f, s);
#ifdef __AVX512DQ__
	v8df dk = _mm512_cvtepi64_pd(k);
#else
	v8si k2 = _mm512_cvtepi64_epi32(k);
	v8df dk = _mm512_cvtepi32_pd(k2);
#endif
	dk = _mm512_mask_add_pd(dk, snmask, dk, m54);
	v8df z = _mm512_mul_pd(s, s);
	v8df R = _mm512_fmadd_pd(log_lg7, z, log_lg6);
	R = _mm512_fmadd_pd(R, z, log_lg5);
	R = _mm512_fmadd_pd(R, z, log_lg4);
	R = _mm512_fmadd_pd(R, z, log_lg3);
	R = _mm512_fmadd_pd(R, z, log_lg2);
	R = _mm512_fmadd_pd(R, z, log_lg1);
	R = _mm512_fmadd_pd(R, z, f);
	R = _mm512_fnmadd_pd(R, s, f);
	v8df y = _mm512_fmadd_pd(dk, ln_two, R);
	// get correct IEEE behavior for arguments <= 0
	__mmask8 zmask = _mm512_cmp_pd_mask(xarg, zero, _CMP_EQ_OS);
	y = _mm512_mask_div_pd(y, zmask, mone, zero);
	__mmask8 nmask = _mm512_cmp_pd_mask(xarg, zero, _CMP_LT_OS);
	y = _mm512_mask_div_pd(y, nmask, zero, zero);
	return y;
}

inline v8df v1log10d(v8df x)
{
	x = v1logd(x);
	return _mm512_div_pd(x, ln_ten);
}

#else

inline v4df v1logd(v4df x)
{
#ifndef VALGRIND
	v4df xarg = x;
#endif
	v4df mask1 = _mm256_castsi256_pd(log_mask1);
	x = _mm256_and_pd(x, mask1);
	v4df snmask = _mm256_cmp_pd(x, dbl_min, _CMP_LT_OS);
	v4df mul = _mm256_blendv_pd(one, two54, snmask);
	x = _mm256_mul_pd(x, mul);
	v4df dkd = _mm256_blendv_pd(zero, m54, snmask);
#ifdef __AVX2__
	v4di ix = _mm256_castpd_si256(x);
	v4di k = _mm256_srli_epi64(ix, 52);
	k = _mm256_sub_epi64(k, log_off);
	v4di iy = _mm256_and_si256(ix, log_mask2);
	v4di iz = _mm256_add_epi64(iy, log_sq2off);
	iz = _mm256_and_si256(iz, log_mask3);
	v4di ie = _mm256_xor_si256(iz, log_mask4);
	iy = _mm256_or_si256(iy, ie);
	iz = _mm256_srli_epi64(iz, 52);
	k = _mm256_add_epi64(k, iz);
	v4di kp = _mm256_permutevar8x32_epi32(k, _mm256_set_epi32(7,5,3,1,6,4,2,0) );
	v4si kk = _mm256_extractf128_si256(kp, 0);
	x = _mm256_castsi256_pd(iy);
	v4df dk = _mm256_cvtepi32_pd(kk);
#else
	v8sf xs =  _mm256_castpd_ps(x);
	xs = _mm256_permute_ps(xs, _MM_SHUFFLE(3, 1, 2, 0));
	v4sf xls = _mm256_extractf128_ps(xs, 0);
	v4si xl = _mm_castps_si128(xls);
	v4sf xhs = _mm256_extractf128_ps(xs, 1);
	v4si xh = _mm_castps_si128(xhs);
	v4si hx = _mm_unpackhi_epi64(xl, xh);
	v4si lx = _mm_unpacklo_epi64(xl, xh);
	v4si k = _mm_srli_epi32(hx, 20);
	k = _mm_sub_epi32(k, log_off);
	hx = _mm_and_si128(hx, log_mask2);
	v4si i = _mm_add_epi32(hx, log_sq2off);
	i = _mm_and_si128(i, log_mask3);
	v4si ii = _mm_xor_si128(i, log_mask4);
	hx = _mm_or_si128(hx, ii);
	v8si xi = _mm256_setzero_si256();
	xl = _mm_unpacklo_epi32(lx,hx);
	xi = _mm256_insertf128_si256(xi, xl, 0);
	xh = _mm_unpackhi_epi32(lx,hx);
	xi = _mm256_insertf128_si256(xi, xh, 1);
	x = _mm256_castsi256_pd(xi);
	i = _mm_srli_epi32(i, 20);
	k = _mm_add_epi32(k, i);
	v4df dk = _mm256_cvtepi32_pd(k);
#endif
	dk = _mm256_add_pd(dk, dkd);
	// f and s have opposite sign compared to openlibm!
	v4df f = _mm256_sub_pd(one, x);
	v4df s = _mm256_sub_pd(two, f);
	s = _mm256_div_pd(f, s);
	v4df z = _mm256_mul_pd(s, s);
#ifdef __FMA__
	v4df R = _mm256_fmadd_pd(log_lg7, z, log_lg6);
	R = _mm256_fmadd_pd(R, z, log_lg5);
	R = _mm256_fmadd_pd(R, z, log_lg4);
	R = _mm256_fmadd_pd(R, z, log_lg3);
	R = _mm256_fmadd_pd(R, z, log_lg2);
	R = _mm256_fmadd_pd(R, z, log_lg1);
	R = _mm256_fmsub_pd(R, z, f);
	R = _mm256_fmsub_pd(R, s, f);
	v4df y = _mm256_fmadd_pd(dk, ln_two, R);
#else
	v4df R = _mm256_mul_pd(log_lg7, z);
	R = _mm256_add_pd(R, log_lg6);
	R = _mm256_mul_pd(R, z);
	R = _mm256_add_pd(R, log_lg5);
	R = _mm256_mul_pd(R, z);
	R = _mm256_add_pd(R, log_lg4);
	R = _mm256_mul_pd(R, z);
	R = _mm256_add_pd(R, log_lg3);
	R = _mm256_mul_pd(R, z);
	R = _mm256_add_pd(R, log_lg2);
	R = _mm256_mul_pd(R, z);
	R = _mm256_add_pd(R, log_lg1);
	R = _mm256_mul_pd(R, z);
	R = _mm256_sub_pd(R, f);
	R = _mm256_mul_pd(R, s);
	R = _mm256_sub_pd(R, f);
	v4df R2 = _mm256_mul_pd(dk, ln_two);
	v4df y = _mm256_add_pd(R, R2);
#endif
	// the following code triggers a bug in valgrind, see
	// https://bugs.kde.org/show_bug.cgi?id=346476
#ifndef VALGRIND
	// get correct IEEE behavior for arguments <= 0
	v4df znmask = _mm256_cmp_pd(xarg, zero, _CMP_LE_OS);
	v4df ieee2 = _mm256_blendv_pd(one, zero, znmask);
	v4df ieee1 = _mm256_sub_pd(ieee2, one);
	v4df nmask = _mm256_cmp_pd(xarg, zero, _CMP_LT_OS);
	ieee1 = _mm256_blendv_pd(ieee1, zero, nmask);
	ieee1 = _mm256_div_pd(ieee1, ieee2);
	y = _mm256_add_pd(y, ieee1);
#endif
	return y;
}

inline v4df v1log10d(v4df x)
{
	x = v1logd(x);
	return _mm256_div_pd(x, ln_ten);
}

#endif // __AVX512F__

VECII_CONST(log_mask1f,0x7fffffff);
#if defined(__AVX512F__) || defined(__AVX2__)
VECII_CONST(log_offf,0x7f);
VECII_CONST(log_mask2f,0x007fffff);
VECII_CONST(log_sq2offf,(0x95f64<<3));
VECII_CONST(log_mask3f,0x00800000);
VECII_CONST(log_mask4f,0x3f800000);
#else
VECI_CONST(log_offf,0x7f);
VECI_CONST(log_mask2f,0x007fffff);
VECI_CONST(log_sq2offf,(0x95f64<<3));
VECI_CONST(log_mask3f,0x00800000);
VECI_CONST(log_mask4f,0x3f800000);
#endif

VECF_CONST(flt_min,FLT_MIN);
VECF_CONST(two25,3.355443200e+07);
VECF_CONST(m25,-25.);

#ifdef __AVX512F__
VECF_CONST(monef,-1.f);
#endif
VECF_CONST(zerof,0.f);
VECF_CONST(onef,1.f);
VECF_CONST(twof,2.f);

VECF_CONST(log_lg4f,-0.24279078841f);
VECF_CONST(log_lg3f,-0.28498786688f);
VECF_CONST(log_lg2f,-0.40000972152f);
VECF_CONST(log_lg1f,-0.66666662693f);

VECF_CONST(ln_twof,logf(2.f));

VECF_CONST(ln_tenf,logf(10.f));

#ifdef __AVX512F__

inline v16sf v1logf(v16sf x)
{
	v16sf xarg = x;
	v16si ix = _mm512_castps_si512(x);
	ix = _mm512_and_epi32(ix, log_mask1f);
	x = _mm512_castsi512_ps(ix);
	__mmask16 snmask = _mm512_cmp_ps_mask(x, flt_min, _CMP_LT_OS);
	x = _mm512_mask_mul_ps(x, snmask, x, two25);
	ix = _mm512_castps_si512(x);
	v16si k = _mm512_srli_epi32(ix, 23);
	k = _mm512_sub_epi32(k, log_offf);
	v16si iy = _mm512_and_epi32(ix, log_mask2f);
	v16si iz = _mm512_add_epi32(iy, log_sq2offf);
	iz = _mm512_and_epi32(iz, log_mask3f);
	v16si ie = _mm512_xor_epi32(iz, log_mask4f);
	iy = _mm512_or_epi32(iy, ie);
	iz = _mm512_srli_epi32(iz, 23);
	k = _mm512_add_epi32(k, iz);
	v16sf f = _mm512_castsi512_ps(iy);
	f = _mm512_sub_ps(f, onef);
	v16sf s = _mm512_add_ps(f, twof);
	s = _mm512_div_ps(f, s);
	v16sf dk = _mm512_cvtepi32_ps(k);
	dk = _mm512_mask_add_ps(dk, snmask, dk, m25);
	v16sf z = _mm512_mul_ps(s, s);
	v16sf R = _mm512_fmadd_ps(log_lg4f, z, log_lg3f);
	R = _mm512_fmadd_ps(R, z, log_lg2f);
	R = _mm512_fmadd_ps(R, z, log_lg1f);
	R = _mm512_fmadd_ps(R, z, f);
	R = _mm512_fnmadd_ps(R, s, f);
	v16sf y = _mm512_fmadd_ps(dk, ln_twof, R);
	// get correct IEEE behavior for arguments <= 0
	__mmask16 zmask = _mm512_cmp_ps_mask(xarg, zerof, _CMP_EQ_OS);
	y = _mm512_mask_div_ps(y, zmask, monef, zerof);
	__mmask16 nmask = _mm512_cmp_ps_mask(xarg, zerof, _CMP_LT_OS);
	y = _mm512_mask_div_ps(y, nmask, zerof, zerof);
	return y;
}

inline v16sf v1log10f(v16sf x)
{
	x = v1logf(x);
	return _mm512_div_ps(x, ln_tenf);
}

#else

inline v8sf v1logf(v8sf x)
{
#ifndef VALGRIND
	v8sf xarg = x;
#endif
	v8sf mask1 = _mm256_castsi256_ps(log_mask1f);
	x = _mm256_and_ps(x, mask1);
	v8sf snmask = _mm256_cmp_ps(x, flt_min, _CMP_LT_OS);
	v8sf mul = _mm256_blendv_ps(onef, two25, snmask);
	x = _mm256_mul_ps(x, mul);
	v8sf dkd = _mm256_blendv_ps(zerof, m25, snmask);
	v8si ix = _mm256_castps_si256(x);
#ifdef __AVX2__
	v8si k = _mm256_srli_epi32(ix, 23);
	k = _mm256_sub_epi32(k, log_offf);
	v8si iy = _mm256_and_si256(ix, log_mask2f);
	v8si iz = _mm256_add_epi32(iy, log_sq2offf);
	iz = _mm256_and_si256(iz, log_mask3f);
	v8si ie = _mm256_xor_si256(iz, log_mask4f);
	iy = _mm256_or_si256(iy, ie);
	iz = _mm256_srli_epi32(iz, 23);
	k = _mm256_add_epi32(k, iz);
#else
	v4si ixl = _mm256_extractf128_si256(ix, 0);
	v4si kl = _mm_srli_epi32(ixl, 23);
	kl = _mm_sub_epi32(kl, log_offf);
	v4si iyl = _mm_and_si128(ixl, log_mask2f);
	v4si iz = _mm_add_epi32(iyl, log_sq2offf);
	iz = _mm_and_si128(iz, log_mask3f);
	v4si ie = _mm_xor_si128(iz, log_mask4f);
	iyl = _mm_or_si128(iyl, ie);
	iz = _mm_srli_epi32(iz, 23);
	kl = _mm_add_epi32(kl, iz);
	v4si ixh = _mm256_extractf128_si256(ix, 1);
	v4si kh = _mm_srli_epi32(ixh, 23);
	kh = _mm_sub_epi32(kh, log_offf);
	v4si iyh = _mm_and_si128(ixh, log_mask2f);
	iz = _mm_add_epi32(iyh, log_sq2offf);
	iz = _mm_and_si128(iz, log_mask3f);
	ie = _mm_xor_si128(iz, log_mask4f);
	iyh = _mm_or_si128(iyh, ie);
	iz = _mm_srli_epi32(iz, 23);
	kh = _mm_add_epi32(kh, iz);
	v8si iy = _mm256_setzero_si256();
	iy = _mm256_insertf128_si256(iy, iyl, 0);
	iy = _mm256_insertf128_si256(iy, iyh, 1);
	v8si k = _mm256_setzero_si256();
	k = _mm256_insertf128_si256(k, kl, 0);
	k = _mm256_insertf128_si256(k, kh, 1);
#endif
	x = _mm256_castsi256_ps(iy);
	v8sf dk = _mm256_cvtepi32_ps(k);
	dk = _mm256_add_ps(dk, dkd);
	// f and s have opposite sign compared to openlibm!
	v8sf f = _mm256_sub_ps(onef, x);
	v8sf s = _mm256_sub_ps(twof, f);
	s = _mm256_div_ps(f, s);
	v8sf z = _mm256_mul_ps(s, s);
#ifdef __FMA__
	v8sf R = _mm256_fmadd_ps(log_lg4f, z, log_lg3f);
	R = _mm256_fmadd_ps(R, z, log_lg2f);
	R = _mm256_fmadd_ps(R, z, log_lg1f);
	R = _mm256_fmsub_ps(R, z, f);
	R = _mm256_fmsub_ps(R, s, f);
	v8sf y = _mm256_fmadd_ps(dk, ln_twof, R);
#else
	v8sf R = _mm256_mul_ps(log_lg4f, z);
	R = _mm256_add_ps(R, log_lg3f);
	R = _mm256_mul_ps(R, z);
	R = _mm256_add_ps(R, log_lg2f);
	R = _mm256_mul_ps(R, z);
	R = _mm256_add_ps(R, log_lg1f);
	R = _mm256_mul_ps(R, z);
	R = _mm256_sub_ps(R, f);
	R = _mm256_mul_ps(R, s);
	R = _mm256_sub_ps(R, f);
	v8sf R2 = _mm256_mul_ps(dk, ln_twof);
	v8sf y = _mm256_add_ps(R, R2);
#endif
	// the following code triggers a bug in valgrind
#ifndef VALGRIND
	// get correct IEEE behavior for arguments <= 0
	v8sf znmask = _mm256_cmp_ps(xarg, zerof, _CMP_LE_OS);
	v8sf ieee2 = _mm256_blendv_ps(onef, zerof, znmask);
	v8sf ieee1 = _mm256_sub_ps(ieee2, onef);
	v8sf nmask = _mm256_cmp_ps(xarg, zerof, _CMP_LT_OS);
	ieee1 = _mm256_blendv_ps(ieee1, zerof, nmask);
	ieee1 = _mm256_div_ps(ieee1, ieee2);
	y = _mm256_add_ps(y, ieee1);
#endif
	return y;
}

inline v8sf v1log10f(v8sf x)
{
	x = v1logf(x);
	return _mm256_div_ps(x, ln_tenf);
}

#endif // __AVX512F__

#else

// stub routines, should never be called
inline int v1logd(int) { return 0; }
inline int v1log10d(int) { return 0; }
inline int v1logf(int) { return 0; }
inline int v1log10f(int) { return 0; }

#endif // __AVX__

// wrapper routines to give math functions C++ linkage
// this prevents warnings from the Oracle Studio compiler
inline double wr_logd(double x)
{
	return log(x);
}

inline sys_float wr_logf(sys_float x)
{
	return logf(x);
}

inline double wr_log10d(double x)
{
	return log10(x);
}

inline sys_float wr_log10f(sys_float x)
{
	return log10f(x);
}


void vlog(const double x[], double y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vlog()" );

	vecfun( x, y, nlo, nhi, wr_logd, v1logd );
}

void vlog10(const double x[], double y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vlog10()" );

	vecfun( x, y, nlo, nhi, wr_log10d, v1log10d );
}

void vlog(const sys_float x[], sys_float y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vlog()" );

	vecfun( x, y, nlo, nhi, wr_logf, v1logf );
}

void vlog10(const sys_float x[], sys_float y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vlog10()" );

	vecfun( x, y, nlo, nhi, wr_log10f, v1log10f );
}

void vlog(double *y, double x0, double x1, double x2, double x3)
{
	V1FUN_PD_4(log, 1.);
}

void vlog10(double *y, double x0, double x1, double x2, double x3)
{
	V1FUN_PD_4(log10, 1.);
}

void vlog(double *y, double x0, double x1, double x2, double x3, double x4, double x5, double x6, double x7)
{
	V1FUN_PD_8(log, 1.);
}

void vlog10(double *y, double x0, double x1, double x2, double x3, double x4, double x5, double x6, double x7)
{
	V1FUN_PD_8(log10, 1.);
}

void vlog(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3)
{
	V1FUN_PS_4(log, 1.f);
}

void vlog10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3)
{
	V1FUN_PS_4(log10, 1.f);
}

void vlog(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	  sys_float x6, sys_float x7)
{
	V1FUN_PS_8(log, 1.f);
}

void vlog10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5,
	    sys_float x6, sys_float x7)
{
	V1FUN_PS_8(log10, 1.f);
}

void vlog(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	  sys_float x6, sys_float x7, sys_float x8, sys_float x9, sys_float x10, sys_float x11, sys_float x12,
	  sys_float x13, sys_float x14, sys_float x15)
{
	V1FUN_PS_16(log, 1.f);
}

void vlog10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	    sys_float x6, sys_float x7, sys_float x8, sys_float x9, sys_float x10, sys_float x11, sys_float x12,
	    sys_float x13, sys_float x14, sys_float x15)
{
	V1FUN_PS_16(log10, 1.f);
}
